--== [ VARIABLES ] ==--
--[ Global variables for RecipeBook ]--
RecipeBookData = {};
--[ Options data ]--
local Realm, Faction, Player;
-- Monitoring status of game tooltip/chatframe
local TooltipUpdated, ChatFrameUpdated = false, false;
-- Debug
local RB_Debug = false;

--[ Hooked Function Variables ]--
local RecipeBook_Old_GameTooltip_OnHide;
local RecipeBook_Old_ContainerFrameItemButton_OnEnter;
local RecipeBook_Old_ContainerFrameItemButton_OnUpdate;
local RecipeBook_Old_AuctionFrameItem_OnEnter;
local RecipeBook_Old_SetItemRef;
local RecipeBook_Old_GameTooltip_SetInventoryItem;
local RecipeBook_Old_GameTooltip_SetMerchantItem;


--== [ ON LOAD ] ==--
function RecipeBook_OnLoad()
	--== Configuration Frame ==--
	UIPanelWindows["RecipeBook_OptionsFrame"] = {area = "center", pushable = 0};
	tinsert(UISpecialFrames, "RecipeBook_OptionsFrame");
	--== Event Registration ==--
	this:RegisterEvent("PLAYER_ENTERING_WORLD"); --For loading data
	this:RegisterEvent("TRADE_SKILL_SHOW");
	this:RegisterEvent("TRADE_SKILL_UPDATE");
	this:RegisterEvent("CRAFT_SHOW");
	this:RegisterEvent("CRAFT_UPDATE"); --These four are the workhorse of the DB
	this:RegisterEvent("AUCTION_ITEM_LIST_UPDATE"); --Auction list updating can otherwise clear tooltips
	this:RegisterEvent("CHAT_MSG_SKILL"); --Skillups for disenchants make baby enchanters happy.
	
	--Slash Command Handlers--
	SlashCmdList["RecipeBook"] = RecipeBook_SlashHandler;
	SLASH_RecipeBook1 = "/recipebook";
	SLASH_RecipeBook2 = "/rb";
	
	DEFAULT_CHAT_FRAME:AddMessage(RECIPEBOOK_LOADED); 

	--==  Function Hooks ==--
	-- Blizzard function hooks --
	RecipeBook_Old_ContainerFrameItemButton_OnEnter = ContainerFrameItemButton_OnEnter; --Good ol' bags
	ContainerFrameItemButton_OnEnter = RecipeBook_ContainerFrameItemButton_OnEnter;
	
	RecipeBook_Old_ContainerFrameItemButton_OnUpdate = ContainerFrameItemButton_OnUpdate; --Bags update!
	ContainerFrameItemButton_OnUpdate = RecipeBook_ContainerFrameItemButton_OnUpdate;

	RecipeBook_Old_AuctionFrameItem_OnEnter = AuctionFrameItem_OnEnter; --Auction Frame
	AuctionFrameItem_OnEnter = RecipeBook_AuctionFrameItem_OnEnter;
	
	RecipeBook_Old_GameTooltip_OnHide = GameTooltip_OnHide;  --Tidy up data
	GameTooltip_OnHide = RecipeBook_GameTooltip_OnHide;
	
	RecipeBook_Old_SetItemRef = SetItemRef; --Chat links
	SetItemRef = RecipeBook_SetItemRef;

	RecipeBook_Old_GameTooltip_SetInventoryItem = GameTooltip.SetInventoryItem; --Bank slots
	GameTooltip.SetInventoryItem = RecipeBook_GameTooltip_SetInventoryItem;
	
	RecipeBook_Old_GameTooltip_SetMerchantItem = GameTooltip.SetMerchantItem; --Bank slots
	GameTooltip.SetMerchantItem = RecipeBook_GameTooltip_SetMerchantItem;
	
	--== Specific hooks for specific mods ==--
	-- My Inventory	--
	if(MyInventoryProfile) then 
		RecipeBook_Old_MIFrameItemButton_OnEnter = MyInventory_FrameItemButton_OnEnter;
		MyInventory_FrameItemButton_OnEnter = RecipeBook_MIFrameItemButton_OnEnter;
		RecipeBook_Old_MIFrameItemButton_OnUpdate = MyInventoryFrameItemButton_OnUpdate;
		MyInventoryFrameItemButton_OnUpdate = RecipeBook_MIFrameItemButton_OnUpdate;
	end
	-- AllInOne Inventory --
	if(AllInOneInventory_ReplaceBags or AllInOneInventory_ReplaceBags == 0) then 
		RecipeBook_Old_AIOI_ModifyItemTooltip = AllInOneInventory_ModifyItemTooltip;
		AllInOneInventory_ModifyItemTooltip = RecipeBook_AIOI_ModifyItemTooltip;
	end
	-- BankStatement --
	if(BankStatementItemButton_OnEnter) then --Checking for BankStatement to play nice.
		RecipeBook_Old_BS_ItemButton_OnEnter = BankStatementItemButton_OnEnter;
		BankStatementItemButton_OnEnter = RecipeBook_BS_ItemButton_OnEnter;
		
		RecipeBook_Old_BS_ContainerFrameItemButton_OnEnter = BankStatementContainerFrameItemButton_OnEnter;
		BankStatementContainerFrameItemButton_OnEnter = RecipeBook_BS_ContainerFrameItemButton_OnEnter;
	end
	-- BankItems --
	if(BankItems_Button_OnEnter) then 
		RecipeBook_Old_BI_Button_OnEnter = BankItems_Button_OnEnter;
		BankItems_Button_OnEnter = RecipeBook_BI_Button_OnEnter;
		
		RecipeBook_Old_BI_BagItem_OnEnter = BankItems_BagItem_OnEnter;
		BankItems_BagItem_OnEnter = RecipeBook_BI_BagItem_OnEnter;
	end
end


--== [ INITIALIZE ] ==--
-- [ InitializeSetup() : New Player? New Data? This makes sure all the options are in the right place. ] --
local function RecipeBook_InitializeSetup()
	Player = UnitName("player");
	Realm = GetRealmName();
	Faction = UnitFactionGroup("player");

	if RecipeBookData == nil then 
		RecipeBookData = {}; 
		RecipeBookData.Version = "1";
	end
	
	--Setting up data structure.
	if RecipeBookData[Realm] == nil then 
		RecipeBookData[Realm] = {};
	end
	if RecipeBookData[Realm][Faction] == nil then 
		RecipeBookData[Realm][Faction] = {};
	end
	
--	This bit is important.  We're changing data structures.  
--	We're going from:
--		Realm
--			Faction
--				Alt
--					Tradeskill
--	to:
--		Realm
--			Faction
--				"Personal"
--					Alt
--						Tradeskill
--				"Shared"
--					Some other player
--						Tradeskill
--	
--	The "Shared" functionality isn't in yet, but it's coming soon and now is as good a time to start using the data structure as any.
--	We're about to do a transparent conversion.  RB 0.5 can NOT be reverted without getting some seriously weird bloaty data tables. 	

	if(RecipeBookData.Version == nil) then 
		local function MoveFactionData(crealm, data)
			--If you have data in this realm's Alliance faction and no "Personal" it's out of date.
			local temptable;
			if((RecipeBookData[crealm]["Alliance"] ~= nil) and (RecipeBookData[crealm]["Alliance"]["Personal"] == nil))then 
				temptable = RecipeBookData[crealm]["Alliance"];
				RecipeBookData[crealm]["Alliance"] = {};
				RecipeBookData[crealm]["Alliance"]["Personal"] = temptable;
			end
			if((RecipeBookData[crealm]["Horde"] ~= nil) and (RecipeBookData[crealm]["Horde"]["Personal"] == nil)) then 
				temptable = RecipeBookData[crealm]["Horde"];
				RecipeBookData[crealm]["Horde"] = {};
				RecipeBookData[crealm]["Horde"]["Personal"] = temptable;
			end
		end
		table.foreach(RecipeBookData, MoveFactionData);
		RecipeBookData.Version = "1";
	end
	
	if RecipeBookData[Realm][Faction]["Personal"] == nil then --Personal Data
		if RecipeBookData[Realm][Faction]~= nil then
			local temptable = RecipeBookData[Realm][Faction];
			RecipeBookData[Realm][Faction] = {};
			RecipeBookData[Realm][Faction]["Personal"] = temptable;
		else 
			RecipeBookData[Realm][Faction]["Personal"] = {};
		end
	end

	
	if RecipeBookData[Realm][Faction]["Personal"][Player] == nil then 
		RecipeBookData[Realm][Faction]["Personal"][Player] = {};
	end  --Your current character.
	if RecipeBookData[Realm][Faction]["Shared"] == nil then 
		RecipeBookData[Realm][Faction]["Shared"] = {};
	end --Data you've been sent.  To be used later.
	if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] ~= nil then
		--Setting up default options should no longer be necessary.  Now converting a few data types.
		-- Output ChatFrame|Tooltip --> Chatframe/Tooltip On|Off
		if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] == "Tooltip" then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Tooltip"] = "On";
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] = nil;
		elseif RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] == "ChatFrame" then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["ChatFrame"] = "On";
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] = nil;
		end
		-- AltData --> AltFaction, do this before SameFaction/OtherFaction conversion
		if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltData"] ~= nil then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] = RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltData"];
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltData"] = nil;
		end --Old data change.
		-- AltFaction Same|Opposite|Both --> SameFaction/OtherFaction On|Off
		if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] == "Same" then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["SameFaction"] = "On"
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["OtherFaction"] = "Off";
		elseif RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] == "Opposite" then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["SameFaction"] = "Off"
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["OtherFaction"] = "On";
		elseif RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] == "Both" then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["SameFaction"] = "On"
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["OtherFaction"] = "On";
		end
	else
		RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] = {};
	end
end




--== [ EVENT HANDLERS ] ==--
-- [ OnEvent(event) : So what happens when an event fires? ] --
function RecipeBook_OnEvent(event)
	if (event == "PLAYER_ENTERING_WORLD") then
		RecipeBook_InitializeSetup();
	elseif(event == "CRAFT_SHOW" or event == "CRAFT_UPDATE") then
		 RecipeBook_UpdateSkillInfo("craft"); --Enchanting window
	elseif(event=="TRADE_SKILL_SHOW" or event == "TRADE_SKILL_UPDATE") then
		 RecipeBook_UpdateSkillInfo("skill"); --All other tradeskills.
	elseif(event=="AUCTION_ITEM_LIST_UPDATE") then
		RecipeBook_DoHookedFunction();
	elseif(event=="CHAT_MSG_SKILL") then
		RecipeBook_ParseChatMessage(arg1);
	end
end

--[ SlashHandler: for those of us who don't quite know how to do nifty GUI stuff yet. ]--
function RecipeBook_SlashHandler(msg)
	local msglist = {};
	local toggle;
	--/rb : Display status.
	if (not msg or msg == "") then 
		RecipeBook_PrintStatus("all");
		return RecipeBook_Print(RECIPEBOOK_USAGE, 2);
	else
		msglist = RecipeBook_Split(string.lower(msg));
	end
	--/rb config : Toggles the config window
	if (msglist[1] == "config") then
		if(RecipeBook_OptionsFrame:IsVisible()) then
			RecipeBook_OptionsFrame:Hide();
		else
			RecipeBook_OptionsFrame:Show();
		end
	--/rb output [chatframe|tooltip|both] [on|off] : Change output location
	elseif(msglist[1] == "output") then 
		if msglist[2] then 
			if msglist[3] then 
				if msglist[3] == "on" then toggle = "On";
				elseif msglist[3] == "off" then toggle = "Off";
				else RecipeBook_Print(RECIPEBOOK_INVALID_OUTPUT, 1);
				end
				if msglist[2] == "chatframe" then 
					RecipeBook_SetOption("ChatFrame", toggle);
				elseif msglist[2] == "tooltip" then 
					RecipeBook_SetOption("Tooltip", toggle);
				elseif msglist[2] == "both" then
					RecipeBook_SetOption("ChatFrame", toggle);
					RecipeBook_SetOption("Tooltip", toggle);
				else 
					RecipeBook_Print(RECIPEBOOK_INVALID_OUTPUT, 1);
				end
			else
				RecipeBook_Print(RECIPEBOOK_INVALID_OUTPUT, 1);
			end
		end
		RecipeBook_PrintStatus("Output");
	--/rb on : Turn RecipeBook on
	elseif(msg == "on") then 
		RecipeBook_SetOption("Status", "On");
		RecipeBook_PrintStatus("Status");
	--/rb off : Turn RecipeBook off
	elseif(msg == "off") then 
		RecipeBook_SetOption("Status", "Off");
		RecipeBook_PrintStatus("Status");
	--/rb status : Display RecipeBook on/off status
	elseif(msg == "status") then  RecipeBook_PrintStatus("Status")
	--/rb version : Display RecipeBook version number
	elseif(msg == "version") then RecipeBook_Print(RECIPEBOOK_VERSION, 0);
	--/rb skill [alt] [tradeskill] Print alt list for a particular tradeskill
	elseif (msglist[1] == "skill") then 
		if(msglist[2] and msglist[3]) then 
			local alt, tradeskill, recipelist = RecipeBook_GetAltData(msglist[2], msglist[3]);
			if alt then return RecipeBook_PrintAltData(alt, tradeskill, recipelist) end;
		elseif not msglist[2] then 
			return RecipeBook_Print(RECIPEBOOK_NOALTGIVEN, 1);
		elseif not msglist[3] then 
			return RecipeBook_Print(RECIPEBOOK_NOTRADESKILLGIVEN, 1);
		end
	--/rb faction [same|opposite|both] : Change factional display
	elseif (msglist[1] == "faction") then
		if msglist[2] then
			if(msglist[2] == "same") then 
				RecipeBook_SetOption("SameFaction", "On");
				RecipeBook_SetOption("OtherFaction", "Off");
			elseif(msglist[2] == "opposite") then 
				RecipeBook_SetOption("SameFaction", "Off");
				RecipeBook_SetOption("OtherFaction", "On");
			elseif(msglist[2] == "both") then 
				RecipeBook_SetOption("SameFaction", "On");
				RecipeBook_SetOption("OtherFaction", "On");
			else 
				RecipeBook_Print(RECIPEBOOK_INVALID_DISPLAY, 1);
			end
		end
		RecipeBook_PrintStatus("AltFaction");
	--/rb debug : Debug toggle
	elseif(msglist[1] == "debug") then
		if (RB_Debug) then 
			RB_Debug = false;
		else 
			RB_Debug = true;
		end
		RecipeBook_Print("Debug: "..(RB_Debug and "ON" or "OFF"), 0);
	--/rb known [on|off] : Show/hide already known
	elseif(msglist[1] == "known") then
		if msglist[2] then 
			if(msglist[2] == "on") then 
				RecipeBook_SetOption("Known", "On");
			elseif(msglist[2] == "off") then 
				RecipeBook_SetOption("Known", "Off");
			else 
				RecipeBook_Print(RECIPEBOOK_INVALID_KNOWN, 1);
			end
		end
		RecipeBook_PrintStatus("Known");
	--/rb learn [on|off] : Show/hide can be learned by
	elseif(msglist[1] == "learn") then
		if msglist[2] then 
			if(msglist[2] == "on") then 
				RecipeBook_SetOption("CanLearn", "On");
			elseif(msglist[2] == "off") then 
				RecipeBook_SetOption("CanLearn", "Off");
			else 
				RecipeBook_Print(RECIPEBOOK_INVALID_CANLEARN, 1);
			end
		end
		RecipeBook_PrintStatus("CanLearn");
	--/rb future [on|off] : Show/hid will be learnable by
	elseif msglist[1] == "future" then
		if msglist[2] then 
			if(msglist[2] == "on") then 
				RecipeBook_SetOption("WillLearn", "On");
			elseif(msglist[2] == "off") then 
				RecipeBook_SetOption("WillLearn", "Off");
			else 
				RecipeBook_Print(RECIPEBOOK_INVALID_WILLLEARN, 1);
			end
		end
		RecipeBook_PrintStatus("WillLearn");
	--/rb specials : Scans for tradeskill specializations.
	elseif msglist[1] == "specials" then
		RecipeBook_SpecialtyScan();
	--Initiate RB Sending Code (not yet working)
	elseif(msglist[1] == "send") then
		RecipeBook_Print("Data send coming soon...");
	--No valid argument given.
	else 
		RecipeBook_Print(RECIPEBOOK_USAGE, 1);
	end
end



--== [ HOOKED FUNCTIONS ] ==--
--[ Vanilla tooltip Updating functions ]--
function RecipeBook_ContainerFrameItemButton_OnEnter()
	RecipeBook_Old_ContainerFrameItemButton_OnEnter();
	RecipeBook_DoHookedFunction();
end
function RecipeBook_AuctionFrameItem_OnEnter(type, index)
	RecipeBook_Old_AuctionFrameItem_OnEnter(type, index);
	RecipeBook_DoHookedFunction();
end
function RecipeBook_SetItemRef(link, button)
	RecipeBook_Old_SetItemRef(link, button);
	TooltipUpdated = false;
	RecipeBook_DoHookedFunction("ItemRefTooltip");
end
--[ GameTooltip:SetInventoryItem() ]--
function RecipeBook_GameTooltip_SetInventoryItem(this, unit, slot) --Specifically a bank hack.
--	if slot < 40 then 
--		return RecipeBook_Old_GameTooltip_SetInventoryItem(this, unit, slot) 
--	end --Not a bank bag.  Die.
	local hasItem, hasCooldown, repairCost = RecipeBook_Old_GameTooltip_SetInventoryItem(this, unit, slot);
	RecipeBook_DoHookedFunction(); 
	return hasItem, hasCooldown, repairCost;
end
--[ GameTooltip:SetMerchantItem() ]--
function RecipeBook_GameTooltip_SetMerchantItem(unit, slot) --Vendors
	RecipeBook_Old_GameTooltip_SetMerchantItem(unit, slot);
	RecipeBook_DoHookedFunction(); 
end
--[ GameTooltip:OnHide() ]--
function RecipeBook_GameTooltip_OnHide()
	TooltipUpdated = false;
	ChatFrameUpdated = false;
	RecipeBook_Old_GameTooltip_OnHide();
end
--[ Container Frame ItemButton OnUpdate Hook ]--
function RecipeBook_ContainerFrameItemButton_OnUpdate(elapsed)
	RecipeBook_Old_ContainerFrameItemButton_OnUpdate(elapsed);
	RecipeBook_ConditionalDoHookedFunction()
end
--==[ My Inventory ]==--
--[ MyInventory FrameItemButton OnEnter Hook ]--
function RecipeBook_MIFrameItemButton_OnEnter() --MyInventory doesn't call the original OnEnter function
	RecipeBook_Old_MIFrameItemButton_OnEnter();
	if (not TooltipUpdated) then 
		RecipeBook_DoHookedFunction();
	end
end
--[ MyInventory FrameItemButton OnUpdate Hook ]--
function RecipeBook_MIFrameItemButton_OnUpdate(elapsed) --MyInventory doesn't call the original OnUpdate function
	RecipeBook_Old_MIFrameItemButton_OnUpdate(elapsed);
	RecipeBook_ConditionalDoHookedFunction();
end
--==[ All-In-One-Inventory ]==--
--[ AIOI Modify Tooltip Hook ]--
function RecipeBook_AIOI_ModifyItemTooltip(bag, slot, tooltipName)
	RecipeBook_Old_AIOI_ModifyItemTooltip(bag, slot, tooltipName);
	RecipeBook_DoHookedFunction();
end
--==[ BankStatement ]==--
--[ BS ItemButton OnEnter Hook ]-- 
function RecipeBook_BS_ItemButton_OnEnter()
	RecipeBook_Old_BS_ItemButton_OnEnter();
	RecipeBook_DoHookedFunction();
end
--[ BS Container ItemButton OnEnter Hook ]--
function RecipeBook_BS_ContainerFrameItemButton_OnEnter()
	RecipeBook_Old_BS_ContainerFrameItemButton_OnEnter();
	RecipeBook_DoHookedFunction();
end
--==[ BankItems ]==--
--[BI Button OnEnter Hook ]--
function RecipeBook_BI_Button_OnEnter()
	RecipeBook_Old_BI_Button_OnEnter();
	RecipeBook_DoHookedFunction();
end
--[ BI BagItem OnEnter Hook ]--
function RecipeBook_BI_BagItem_OnEnter()
	RecipeBook_Old_BI_BagItem_OnEnter();
	RecipeBook_DoHookedFunction();
end

function RecipeBook_ConditionalDoHookedFunction(tooltip)
	if not tooltip then tooltip = "GameTooltip" end;
	TooltipUpdated = false;
	local index;
	for index = GameTooltip:NumLines(), 3, -1 do
		local text = getglobal("GameTooltipTextLeft"..index):GetText();
		if text ~= nil and (string.find(text, RECIPEBOOK_KNOWNBY) or string.find(text, RECIPEBOOK_CANLEARN) or string.find(text, RECIPEBOOK_WILLLEARN)) then 
			TooltipUpdated = true;
			break;
		end
	end
	if not TooltipUpdated then 
		RecipeBook_DoHookedFunction(); 
	end
end

--== [ DO THE HOOKED FUNCTION ] ==--
function RecipeBook_DoHookedFunction(tooltip)
	ItemTitle = nil;
	if RecipeBook_GetOption("Status") then 
		if not tooltip then 
			tooltip = "GameTooltip";
		end
		if(TooltipUpdated) then --only one mod per tooltip!
			return;
		end
		local text = getglobal(tooltip.."TextLeft1"):GetText();
		if text ~= nil then 
			local recipename = nil;
			local recipename, tradeskill, skillevel, specialty = RecipeBook_GetRecipeData(tooltip, text);  --Let's start at the very beginning.
			local altsknow, altslearn, altshaveskill;
			if(recipename ~= nil) then 
				altsknow, altslearn, altshaveskill = RecipeBook_MatchRecipeData(recipename, tradeskill, skillevel, specialty);
 				if RecipeBook_GetOption("Tooltip") then
					RecipeBook_OutputRecipeTooltip(tooltip, altsknow, altslearn, altshaveskill);
				end
				if RecipeBook_GetOption("ChatFrame") and not ChatFrameUpdated then
					RecipeBook_OutputRecipeChatFrame(recipename, altsknow, altslearn, altshaveskill);
				end
			end
		end
	end
end



--== [ HANDLING TRADESKILL UPDATES ] ==--
--[ UpdateSkillInfo : When the Tradeskill/Craft window opens, this wipes and replaces the known recipes list for that player. ]--
function RecipeBook_UpdateSkillInfo(skilltype)
	local skillrank,maxranks,total;
	if(skilltype == "craft") then
		total = GetNumCrafts();
		skill, skillrank, maxranks = GetCraftDisplaySkillLine();
	else
		total = GetNumTradeSkills();
		skill, skillrank, maxranks = GetTradeSkillLine();
	end
	if skill == nil then 
		return false;
	end --Apparently hunter skills trigger the Tradeskill/Craft window too.
	local tradeskilltable = {};
	if(RecipeBookData[Realm][Faction]["Personal"][Player][skill]) then
		tradeskilltable.Specialty = RecipeBookData[Realm][Faction]["Personal"][Player][skill].Specialty;
		tradeskilltable.SubSpecialty = RecipeBookData[Realm][Faction]["Personal"][Player][skill].SubSpecialty;
	end
	
--	RecipeBookData[Realm][Faction]["Personal"][Player][skill] = {}; 
	tradeskilltable.Rank = skillrank;
	local name,temp,crafttype;
	for index = 0, total, 1 do
		if(skilltype == "craft") then 
			name, temp, crafttype, temp, temp = GetCraftInfo(index);
		else 
			name, crafttype, temp, temp = GetTradeSkillInfo(index);
		end
		if((crafttype ~= "header") and (name ~= nil)) then
			table.insert(tradeskilltable, name);
		end
	end
	if(tradeskilltable ~= nil and RecipeBookData[Realm][Faction]["Personal"][Player][skill]~= nil) then
		if(table.getn(tradeskilltable) > table.getn(RecipeBookData[Realm][Faction]["Personal"][Player][skill])) then
			RecipeBookData[Realm][Faction]["Personal"][Player][skill] = tradeskilltable;
		else
			RecipeBookData[Realm][Faction]["Personal"][Player][skill].Rank = skillrank;
		end
	elseif(tradeskilltable ~= nil) then 
		RecipeBookData[Realm][Faction]["Personal"][Player][skill] = tradeskilltable;
	end
end
--[ ParseChatMessage : When you get a skillup, update the tradeskill rank only. ]--
function RecipeBook_ParseChatMessage(text)
	local skill, rank;
	gsub(text, "Your skill in (%w*) has increased to (%d+)%.", function(a,b) skill = a; rank = b; end);
	if(RecipeBookData[Realm][Faction]["Personal"][Player][tradeskill] ~= nil) then 
		(RecipeBookData[Realm][Faction]["Personal"][Player][tradeskill]).Rank = rank;
	end
end
--[ SpecialtyScan : /rb specials will scan your skills for tradeskill specializations. ]--
function RecipeBook_SpecialtyScan()
	local SpecString = "None";
	local SpecList = {};
	
	local name, texture, offset, numSpells = GetSpellTabInfo(1);
	for index = 1, numSpells, 1 do 
		local spellName, spellRank = GetSpellName(index, BOOKTYPE_SPELL);
		if(spellName) then
			local spellfind = "|"..spellName.." |";
			if string.find(RECIPEBOOK_BLACKSMITHING_SPECIALS, spellfind) then 
				if(RecipeBookData[Realm][Faction]["Personal"][Player]["Blacksmithing"] == nil) then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Blacksmithing"] = {};
				end
				if(string.find(spellName, "Master")) then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Blacksmithing"].Specialty = "Weaponsmith";
					RecipeBookData[Realm][Faction]["Personal"][Player]["Blacksmithing"].SubSpecialty = spellName;
					table.insert(SpecList, "Weaponsmith ("..spellName..")");
				else
					RecipeBookData[Realm][Faction]["Personal"][Player]["Blacksmithing"].Specialty = spellName;
					table.insert(SpecList, spellName);
				end
			elseif string.find(RECIPEBOOK_ENGINEERING_SPECIALS, spellfind) then 
				if(RecipeBookData[Realm][Faction]["Personal"][Player]["Engineering"] == nil) then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Engineering"] = {};
				end
				RecipeBookData[Realm][Faction]["Personal"][Player]["Engineering"].Specialty = spellName;
				table.insert(SpecList, spellName);
			elseif string.find(RECIPEBOOK_LEATHERWORKING_SPECIALS, spellfind) then 
				if(RecipeBookData[Realm][Faction]["Personal"][Player]["Leatherworking"] == nil) then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Leatherworking"] = {};
				end
				RecipeBookData[Realm][Faction]["Personal"][Player]["Leatherworking"].Specialty = spellName;
				table.insert(SpecList, spellName);
			end
		end
		if table.getn(SpecList) > 0 then SpecString = table.concat(SpecList, ", ") end;
	end
	RecipeBook_Print("Found and added the following tradeskill specializations: "..SpecString, 0);
end


--== [ CHECK THE TOOLTIP DATA ] ==--
function RecipeBook_GetRecipeData(tooltip, text)
	local recipename, tradeskill, skillevel, specialty;
	-- Take tooltip line 1, look for "X: Y".  If X is in the list of pattern names, Y is the name of the pattern.
	if string.find("Formula Pattern Plans Recipe Schematic", string.gsub(text, "(%w+)%: (.+)", function(a,b) recipename = b; return a end)) then -- It's a recipe!
		-- Take tooltip line 2, look for "Requires X (#)" and return X and # if found.  Otherwise this will return "text" in toto, meaning it's a malformed tradeskill requirement.
		text = getglobal(tooltip.."TextLeft2"):GetText();
		if(text ~= nil) then 
			string.gsub(text, "Requires (%w+) %((%d+)%)", function(a,b) tradeskill = a; skillevel = b end); 
		else 
			return nil;
		end -- This line shouldn' be empty, but just in case.
		if tradeskill == text then 
			return nil;
		end --No tradeskill requirement, no go.
		-- Take tooltip line 3, look for "Requires XY", where X is not "Level" and Y doesn't have to exist (i.e. "Requires Armorsmithing" only has X), and return XY.
		text = getglobal(tooltip.."TextLeft3"):GetText();
		if(text ~= nil) then 
			if(string.find(text, "Requires")) then 
				specialty = string.gsub(text, "(%w+)%s*(%w+)(.*)", function(a,b,c) if a == "Requires" and not (b == "Level") then return b..c else return false end end)
			else
				specialty = false;
			end
		else 
			specialty = false;
		end
		if text == specialty then 
			specialty =  false;
		end --No specialty requirement, that's cool.
		-- Put it all together, and what do you got?
		return recipename, tradeskill, tonumber(skillevel), specialty;
	else 
		return nil;  -- It's not a recipe!
	end
end


--== [ CHECK THE ALT DATA ] ==--
-- [ MatchRecipeData(recipename, tradeskill, rank, specialty) : Given tradeskill info, calls FindInFaction on each appropriate faction, then colour-codes appropriately. ] --
function RecipeBook_MatchRecipeData(recipename, tradeskill, rank, specialty)
	local knows, canlearn, hastradeskill;
	local otherfaction = (Faction == "Alliance") and "Horde" or "Alliance";
	local function tablestostring(table1, table2)
		local tstring = nil;
		table1 = table1 and table1 or {};
		table2 = table2 and table2 or {};
		if(table.getn(table1) > 0) then 
			tstring = table.concat(table1, ", ");
		end
		if(table.getn(table2) > 0) then
			if tstring then 
				tstring = tstring..", "..table.concat(table2, ", ");
			else 
				tstring = table.concat(table2, ", ");
			end
		end
		return tstring;
	end
	if RecipeBook_GetOption("SameFaction") then 
		if RecipeBook_GetOption("OtherFaction") then 
			if RecipeBookData[Realm][Faction] ~= nil then 
				knows, canlearn, hastradeskill = RecipeBook_FindInFaction(Faction, recipename, tradeskill, rank, specialty);
			end
			if RecipeBookData[Realm][otherfaction] ~= nil then 
				local oknows, ocanlearn, ohastradeskill;
				oknows, ocanlearn, ohastradeskill = RecipeBook_FindInFaction(otherfaction, recipename, tradeskill, rank, specialty);
			end
			return tablestostring(knows, oknows), tablestostring(canlearn, ocanlearn), tablestostring(hastradeskill, ohastradeskill);
		else
			if RecipeBookData[Realm][Faction] ~= nil then 
				knows, canlearn, hastradeskill = RecipeBook_FindInFaction(Faction, recipename, tradeskill, rank, specialty);
				return table.concat(knows, ", "),  table.concat(canlearn, ", "), table.concat(hastradeskill, ", ");
			else
				return nil;
			end
		end
	elseif RecipeBook_GetOption("OtherFaction") then
		if RecipeBookData[Realm][otherfaction] ~= nil then 
			knows, canlearn, hastradeskill = RecipeBook_FindInFaction(otherfaction, recipename, tradeskill, rank, specialty);
			return table.concat(knows, ", "),  table.concat(canlearn, ", "), table.concat(hastradeskill, ", ");
		else
			return nil;
		end
	end
end

-- [ FindInFaction(pass faction, recipename, tradeskill, rank, specialty) : Returns three lists of who knows, can learn, and has tradeskill for the passed faction. ] --
function RecipeBook_FindInFaction(lfaction, recipename, tradeskill, rank, specialty)
	local knows = {};
	local canlearn = {};
	local hastradeskill = {};
	local color = (lfaction == Faction) and RECIPEBOOK_FACTION_COLOR or RECIPEBOOK_OPPOSITE_COLOR;
	-- Matching recipes (see call).
	local function matchrecipe(who, skilltable)
		local prank, pspec, psubspec, pmatch;
		-- For each recipe in [Tradeskill][Recipe Table] if the recipe name matches then return true, else return nil.
		local function knowsrecipe (tableskill, recipetable)
			if tableskill == tradeskill then 
				pmatch = false;
				for index = 1, table.getn(recipetable), 1 do
					if(recipetable[index] == recipename) then 
						pmatch = true;
						break;
					end
				end
				prank = recipetable.Rank;
				pspec = recipetable.Specialty;
				psubspec = recipetable.SubSpecialty;
				return true;
			else
				return nil;
			end
		end
		
		-- Skip the current alt
		if who == Player then 
			return nil;
		end
		-- For each tradeskill in [Player]'s tradeskill table, if the tradeskill matches then return tradeskill rank, tradeskill specialty, and parse the list of known recipes - else nil.
		table.foreach(skilltable, knowsrecipe);
		-- If a tradeskill match is found:
		if(prank) then 
			who = color..who;
			-- If it's known, insert into the known list.
			if pmatch then 
					table.insert(knows, who..RECIPEBOOK_END_COLOR);
			-- If a specialty is required for this recipe, check to make sure specialty is known.  If it's not, then it's not learnable.
			elseif specialty then 
				if (pspec and pspec == specialty) or (psubspec and psubspec == specialty) then 
					if prank >= rank then 
						table.insert(canlearn, who..RECIPEBOOK_END_COLOR); 
					else 
						table.insert(hastradeskill, who..color.." ("..prank..")"..RECIPEBOOK_END_COLOR);
					end
				else
				end
			-- No specialty required; matches already listed, just checking to see if it's learnable now.
			else
				if prank >= rank then 
					table.insert(canlearn, who..RECIPEBOOK_END_COLOR); 
				else 
					table.insert(hastradeskill, who..color.." ("..prank..")"..RECIPEBOOK_END_COLOR);
				end
			end
		-- No tradeskill match was returned.
		else
			return nil;
		end
	end
	
	if(RecipeBookData[Realm] and RecipeBookData[Realm][lfaction] and RecipeBookData[Realm][lfaction]["Personal"]) then 
		table.foreach(RecipeBookData[Realm][lfaction]["Personal"], matchrecipe);
	end
	return knows, canlearn, hastradeskill;
end


--== [ ALT TRADESKILL LISTING ] ==--
--[ GetAltData(who, what) : Finds the appropriate tradeskill table for a given alt and tradeskill. ]--
function RecipeBook_GetAltData(who, what)
	who = string.gsub(who, "%l", string.upper, 1 );
	what = string.gsub(what, "%l", string.upper, 1);
	if RecipeBookData[Realm][Faction]["Personal"][who] then
		local alttable = RecipeBookData[Realm][Faction]["Personal"][who][what];
		if (alttable ~= nil)  then
			return who, what, alttable;
		else
			RecipeBook_Print(RECIPEBOOK_NOTRADESKILLMATCH..'"'..what..'".', 1);
			return false;
		end
	else
		RecipeBook_Print(RECIPEBOOK_NOALTMATCH..'"'..who..'".', 1);
		return false;
	end
end
		
--[ PrintAltData(alt, tradeskill, recipelist) : formats and outputs the data.  Currently to RecipeBook chat tab. ]--
function RecipeBook_PrintAltData(alt, tradeskill, recipelist)
	local chatframe = RecipeBook_GetChatFrame();
	local specialty = "No specialty";
	if(recipelist.Specialty) then 
		specialty = recipelist.Specialty;
		if recipelist.SubSpecialty then specialty = specialty.." ["..recipelist.SubSpecialty.."]" end;
	end
	local rank = recipelist["Rank"] and recipelist["Rank"] or 0;
	local length = table.getn(recipelist);
	RecipeBook_Print(RECIPEBOOK_OUTPUT_CHATFRAME, 0);
	RecipeBook_Print(RECIPEBOOK_PRINTTRADESKILLS..alt..":", -1, chatframe);
	RecipeBook_Print("=== Tradeskill : "..tradeskill.." (Specialty: "..specialty..") ===", -1, chatframe);
	RecipeBook_Print("=== "..length.." recipes known (current skill "..(rank)..") ===", -1, chatframe);
	for index=1, length, 1 do
		if(recipelist[index] ~= nil) then 
			RecipeBook_Print(recipelist[index], -1, chatframe);
		end
	end
	return true;
end

--== [ OUTPUT TO TOOLTIP OR CHAT FRAME ] ==--
function RecipeBook_OutputRecipeTooltip(tooltip, knows, canlearn, havetradeskill)
	local modded = false;
	if tooltip == "ItemRefTooltip" then 
		tooltip = ItemRefTooltip;
	else 
		tooltip = GameTooltip;
	end
	if(RecipeBook_GetOption("Known") and knows) then
		if not modded then 
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_KNOWNBY..knows, 1,0,0);
		modded = true;
	end
	if(RecipeBook_GetOption("CanLearn") and canlearn) then
		if not modded then 
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_CANLEARN..canlearn, 0,1,0);
		modded = true;
	end
	if(RecipeBook_GetOption("WillLearn") and havetradeskill) then
		if not modded then 
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_WILLLEARN..havetradeskill, 0,0,1);
		modded = true;
	end
	tooltip:Show();
	TooltipUpdated = true;
end

function RecipeBook_OutputRecipeChatFrame(recipename, knows, canlearn, havetradeskill)
	local chatframe = RecipeBook_GetChatFrame();
	if RecipeBook_GetOption("Known") then 
		if knows then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_KNOWNBY .. knows, -1, chatframe, 1,0,0);
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NONEKNOWN, -1, chatframe, 0,1,0);
		end
	end
	if RecipeBook_GetOption("CanLearn") then 
		if(canlearn) then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_CANLEARN .. canlearn, -1, chatframe, 1,0,0);
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NONELEARN, -1, chatframe, 0,1,0);
		end
	end
	if RecipeBook_GetOption("WillLearn") then
		if(havetradeskill) then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_WILLLEARN ..havetradeskill, -1, chatframe, 1,0,0);
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NONEWILLLEARN, -1, chatframe, 0,1,0);
		end
	end
	ChatFrameUpdated = true;
end



--== [ UTILITY FUNCTIONS ] ==--
--[ Split : Splits a string into a table of words. ]--
function RecipeBook_Split(tocut)
	local words = {};
	function towords(word) table.insert(words, word) end
	if not string.find(string.gsub(tocut, "%S+", towords), "%S") then 
		return words;
	end
end

--[ Status(message) : Prints a RecipeBook status message. ]--
function RecipeBook_Print(message, prefix, frame, r,g,b)
	if frame == nil then frame = DEFAULT_CHAT_FRAME end;
	if r == nil then r = 1; g = 1; b = 1;
	elseif g == nil then g = 1; b = 1;
	elseif b == nil then b = 1;
	end

	if(prefix == 0) then prefix = RECIPEBOOK_PREFIX;
	elseif(prefix == 1) then prefix = RECIPEBOOK_ERROR_PREFIX;
	else prefix = "";
	end
	frame:AddMessage(prefix..message, r, g, b);
end
--[ Debug(message) : Prints a debug message if Debug is on. ]--
function RecipeBook_Debug(message)
	if(RB_Debug) then
		DEFAULT_CHAT_FRAME:AddMessage("<RecipeBook Debug:> "..message);
	end
end

--[ PrintStatus : Prints out the status of whatever options are passed to it. ]--
function RecipeBook_PrintStatus(option)
	if(option == "all") then --Print all options
		RecipeBook_PrintStatus("Status");
		RecipeBook_PrintStatus("Output");
		RecipeBook_PrintStatus("AltFaction");
		RecipeBook_PrintStatus("Known");
		RecipeBook_PrintStatus("CanLearn");
		RecipeBook_PrintStatus("WillLearn");
	elseif(option == "Output") then --Print output status
		RecipeBook_Print(RECIPEBOOK_OUTPUT_TOOLTIP..(RecipeBook_GetOption("Tooltip") and RECIPEBOOK_ON or RECIPEBOOK_OFF), -1)
		RecipeBook_Print(RECIPEBOOK_OUTPUT_CHATFRAME..(RecipeBook_GetOption("ChatFrame") and RECIPEBOOK_ON or RECIPEBOOK_OFF), -1)
	elseif(option == "Status") then --Print on/off status
		if RecipeBook_GetOption("Status") then
			RecipeBook_Print(RECIPEBOOK_STATUS_ON, 0);
		else
			RecipeBook_Print(RECIPEBOOK_STATUS_OFF, 0);
		end
	elseif(option == "AltFaction") then --Print factional display status
		if RecipeBook_GetOption("SameFaction") then 
			if RecipeBook_GetOption("OtherFaction") then
				RecipeBook_Print(RECIPEBOOK_DISPLAY_BOTHFACTIONS, -1);
			else
				RecipeBook_Print(string.format(RECIPEBOOK_DISPLAY_FACTION, Faction), -1);
			end
		elseif RecipeBook_GetOption("OtherFaction") then
			if(Faction == "Alliance") then 
				RecipeBook_Print(string.format(RECIPEBOOK_DISPLAY_FACTION, "Horde"), -1);
			else 
				RecipeBook_Print(string.format(RECIPEBOOK_DISPLAY_FACTION, "Alliance"), -1);
			end
		else
			RecipeBook_Print(RECIPEBOOK_DISPLAY_NOFACTIONS, -1);
		end
	elseif (option == "Known") then 
		if  RecipeBook_GetOption("Known") then
			RecipeBook_Print(RECIPEBOOK_KNOWN_ON, -1);
		else
			RecipeBook_Print(RECIPEBOOK_KNOWN_OFF, -1);
		end
	elseif(option == "CanLearn") then
		if RecipeBook_GetOption("CanLearn") then
			RecipeBook_Print(RECIPEBOOK_CANLEARN_ON, -1);
		else
			RecipeBook_Print(RECIPEBOOK_CANLEARN_OFF, -1);
		end
	elseif(option == "WillLearn") then 
		if RecipeBook_GetOption("WillLearn") then
			RecipeBook_Print(RECIPEBOOK_WILLLEARN_ON, -1);
		else
			RecipeBook_Print(RECIPEBOOK_WILLLEARN_OFF, -1);
		end
	end
end

--[ GetChatFrame (courtesy Averice's WIM, used by permission): Finds out which Chat Frame the RecipeBook info is in ]--
function RecipeBook_GetChatFrame()
	local frame = nil;
	local frametab, tempframe;
	for i = 1, NUM_CHAT_WINDOWS do
		tempframe = getglobal("ChatFrame" .. i .. "Tab");
		if (tempframe:GetText() == "Recipe Book") then
			frame = getglobal("ChatFrame" .. i);
			frametab = tempframe;
			break;
		end
	end
	if (frame) then
		if(not frametab:IsVisible()) then 
			frametab:Show(); 
		end
		return frame;
	else
		FCF_OpenNewWindow("Recipe Book");
		frame = RecipeBook_GetChatFrame(); 
		FCF_SetLocked(frame, true, false);
		ChatFrame_RemoveAllMessageGroups(frame);
		return frame;
	end
end

--== [ OPTIONS MODIFICATION FUNCTIONS ]==--
--[ UpdateOptions(options to set) : Updates the whole set of options at once. NOT idiot-proof.]--
function RecipeBook_UpdateOptions(newoptions)
	if(newoptions ~= nil) then 
		table.foreach(newoptions, RecipeBook_SetOption);
	end
end
--[ GetOption(option) : Returns the value of a given option; default if no value ]--
function RecipeBook_GetOption(option)
	if(option == "all") then 
		return RecipeBookData[Realm][Faction]["Personal"][Player]["Options"];
	elseif string.find(RECIPEBOOK_VALID_OPTIONS, option) then 
		if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option] then
			option = RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option];
		else
			option = RECIPEBOOK_DEFAULT_OPTIONS[option];
		end
		if option == "On" then 
			return true;
		elseif option == "Off" then 
			return false;
		else 
			return option;
		end
	else
		return nil;
	end
end
--[ SetOption(option) : Sets the value of a given option.  Nil to reset to default. ]--
function RecipeBook_SetOption(option, value)
	if ((type(option) == "string") and (string.find(RECIPEBOOK_VALID_OPTIONS, option))) then
		if (value ~= nil) then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option] = value;
		else
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option] = RECIPEBOOK_DEFAULT_OPTIONS[option];
		end
	end
end
--[ SetDefaultOptions() : Resets options, 'cause you may need it. ]--
function RecipeBook_SetDefaultOptions()
	RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] = {};
end
